// TEcho_APRS.ino

// Programme de réception des trames APRS avec un module LILYGO T-ECHO.
// Adapté par Frank LOBA (www.f4hdb.fr).
// Date création : 11/02/2024.
// Dernière mise à jour : 12/02/2024.

// Programm zum Empfang von APRS-Frames mit einem LILYGO T-ECHO-Modul.
// Adaptiert von Frank LOBA (www.f4hdb.fr).
// Erstellungsdatum: 11.02.2024.
// Letzte Aktualisierung: 12.02.2024.

// Program for receiving APRS frames with a LILYGO T-ECHO module.
// Adapted by Frank LOBA (www.f4hdb.fr).
// Creation date: 02/11/2024.
// Last update: 02/12/2024.


/**** INCLUSION DES BIBLIOTHEQUES GENERALES ****/
/**** EINBEZIEHUNG ALLGEMEINER BIBLIOTHEKEN ****/
/**** INCLUSION OF GENERAL LIBRARIES ****/
#include <Arduino.h>
#include <SPI.h>


/**** INCLUSION DES BIBLIOTHEQUES NECESSAIRES AU PILOTAGE DE L'ECRAN E-PAPER ****/
/**** EINSCHLUSS VON BIBLIOTHEKS, DIE ZUR STEUERUNG DES E-PAPER-BILDSCHIRMS ERFORDERLICH SIND ****/
/**** INCLUSION OF LIBRARIES REQUIRED TO CONTROL THE E-PAPER SCREEN ****/
#include <GxEPD.h>
#include <GxDEPG0150BN/GxDEPG0150BN.h>  // 1.54" b/w
#include <Fonts/FreeMonoBold9pt7b.h>
#include <GxIO/GxIO_SPI/GxIO_SPI.h>
#include <GxIO/GxIO.h>


/**** INCLUSION DES BIBLIOTHEQUES NECESSAIRES AU PILOTAGE DE LA RADIO LORA ****/
/**** EINBEZIEHUNG DER ZUR STEUERUNG DES LORA-RADIOS NOTWENDIGEN BIBLIOTHEKEN ****/
/**** INCLUSION OF THE LIBRARIES NECESSARY TO CONTROL THE LORA RADIO ****/
#include <RadioLib.h>


/**** DECLARATION DE LA MACRO DE CALCUL DU NUMERO DE LA BROCHE ****/
/**** ERKLÄRUNG DES PIN-NUMMER-BERECHNUNGSMAKROS ****/
/**** DECLARATION OF THE PIN NUMBER CALCULATION MACRO ****/
#ifndef _PIN_NUMBER
#define _PIN_NUMBER(_PORT, _PIN)    ((_PORT << 5) + _PIN)  // = Port * 32 + Pin.
#endif


/**** DECLARATION DES NUMEROS DES BROCHES DE L'ECRAN E-PAPER ****/
/**** ERKLÄRUNG DER E-PAPER-BILDSCHIRM-PIN-NUMMERN ****/
/**** DECLARATION OF E-PAPER SCREEN PIN NUMBERS ****/
#define E_PAPER_MISO      _PIN_NUMBER(1,6)
#define E_PAPER_MOSI      _PIN_NUMBER(0,29)
#define E_PAPER_SCLK      _PIN_NUMBER(0,31)
#define E_PAPER_CS        _PIN_NUMBER(0,30)
#define E_PAPER_DC        _PIN_NUMBER(0,28)
#define E_PAPER_RST       _PIN_NUMBER(0,2)
#define E_PAPER_BUSY      _PIN_NUMBER(0,3)
#define E_PAPER_BACKLIGHT _PIN_NUMBER(1,11)


/**** DECLARATION DES PARAMETRES RADIO LORA ****/
/**** ERKLÄRUNG DER LORA-FUNKPARAMETER ****/
/**** DECLARATION OF LORA RADIO PARAMETERS ****/
#define LORA_MISO         _PIN_NUMBER(0,23)
#define LORA_MOSI         _PIN_NUMBER(0,22)
#define LORA_SCLK         _PIN_NUMBER(0,19)
#define LORA_CS           _PIN_NUMBER(0,24)
#define LORA_RST          _PIN_NUMBER(0,25)
#define LORA_DI0          _PIN_NUMBER(0,22)
#define LORA_DIO1         _PIN_NUMBER(0,20)
#define LORA_DIO3         _PIN_NUMBER(0,21)
#define LORA_BUSY         _PIN_NUMBER(0,17)

#define LORA_FREQUENCY        433.775
#define LORA_SPREADING_FACTOR 12
#define LORA_SIGNAL_BANDWIDTH 125.0
#define LORA_CODING_RATE_4    5
#define LORA_SYNC_WORD        0x12
#define LORA_TX_POWER_DBM     20

#define LORA_START_BYTE_1 '<'
#define LORA_START_BYTE_2 0xFF
#define LORA_START_BYTE_3 0x01


/**** DECLARATION DES PARAMETRES SYSTEME ****/
/**** ERKLÄRUNG DER SYSTEMPARAMETER ****/
/**** DECLARATION OF SYSTEM PARAMETERS ****/
#define USER_TOUCH          _PIN_NUMBER(0,11)
#define USER_BUTTON         _PIN_NUMBER(1,10)
#define POWER_ENABLE        _PIN_NUMBER(0,12)


/**** DECLARATION DES OBJETS NECESSAIRES POUR PILOTER L'ECRAN ****/
/**** ERKLÄRUNG DER ZUR STEUERUNG DES BILDSCHIRMS ERFORDERLICHEN OBJEKTE ****/
/**** DECLARATION OF OBJECTS NECESSARY TO CONTROL THE SCREEN ****/
SPIClass    DisplayPort(NRF_SPIM2, E_PAPER_MISO, E_PAPER_SCLK, E_PAPER_MOSI);
GxIO_Class  Io(DisplayPort, E_PAPER_CS, E_PAPER_DC, E_PAPER_RST);
GxEPD_Class Display(Io, E_PAPER_RST, E_PAPER_BUSY);


/**** DECLARATION DES OBJETS ET VARIABLES NECESSAIRES POUR PILOTER LA RADIO LORA ****/
/**** ERKLÄRUNG DER OBJEKTE UND VARIABLEN, DIE ZUR STEUERUNG DES LORA-RADIOS ERFORDERLICH SIND ****/
/**** DECLARATION OF OBJECTS AND VARIABLES NECESSARY TO CONTROL THE LORA RADIO ****/
SPIClass    RfPort(NRF_SPIM3, LORA_MISO, LORA_SCLK, LORA_MOSI);
SPISettings SpiSettings; 
SX1262      Radio = nullptr;

volatile bool Receive = false;
String        TNC2Frame = "";
int           State;


/**** DECLARATION DES OBJETS ET VARIABLES GLOBALES ****/
/**** DEKLARATION VON OBJEKTEN UND GLOBALEN VARIABLEN ****/
/**** DECLARATION OF OBJECTS AND GLOBAL VARIABLES ****/
bool  PowerEnable = false;


/**** DEFINITION DE LA FONCTION DE L'INTERRUPTION POUR LA RECEPTION RADIO LORA ****/
/**** DEFINITION DER UNTERBRECHUNGSFUNKTION FÜR DEN LORA-RADIOEMPFANG ****/
/**** DEFINITION OF THE INTERRUPTION FUNCTION FOR LORA RADIO RECEPTION ****/
void SetRxFlag(void)
{
  Receive = true;
}


/**** DEFINITION DE LA FONCTION D'INITIALISATION DU MODULE LILYGO T-ECHO ****/
/**** DEFINITION DER INITIALISIERUNGSFUNKTION DES LILYGO T-ECHO-MODULS ****/
/**** DEFINITION OF THE INITIALIZATION FUNCTION OF THE LILYGO T-ECHO MODULE ****/
void setup()
{
    // Attente.
    // Warten.
    // Wait.
    delay(1500);

    // Affectation des entrées et sorties.
    // Zuordnung der Ein- und Ausgänge.
    // Assignment of inputs and outputs.
    pinMode(POWER_ENABLE, OUTPUT);
    pinMode(E_PAPER_BACKLIGHT, OUTPUT);

    pinMode(USER_BUTTON, INPUT_PULLUP);
    pinMode(USER_TOUCH, INPUT_PULLUP);

    // Mise en marche.
    // Ein.
    // Power on.
    digitalWrite(POWER_ENABLE, HIGH);
    PowerEnable = true;

    // Initialisation des objets de l'écran.
    // Initialisierung von Bildschirmobjekten.
    // Initialization of screen objects.
    DisplayPort.begin();
    Display.init(/*115200*/);

    // Rotation de l'écran.
    // Bildschirm drehen.
    // Rotate the screen.
    Display.setRotation(3);

    // Remplissage de l'écran avec la couleur blanche (pour l'effacer).
    // Den Bildschirm mit weißer Farbe füllen (um ihn zu löschen).
    // Filling the screen with white color (to clear it).
    Display.fillScreen(GxEPD_WHITE);

    // Attribution de la couleur noir au texte affiché à l'écran.
    // Dem auf dem Bildschirm angezeigten Text die Farbe Schwarz zuweisen.
    // Assigning the color black to the text displayed on the screen.
    Display.setTextColor(GxEPD_BLACK);

    // Attribution de la police de caractères affichés à l'écran.
    // Zuweisung der auf dem Bildschirm angezeigten Schriftart.
    // Assignment of the font displayed on the screen.
    Display.setFont(&FreeMonoBold9pt7b);

    // Positionnement du curseur pour écrire à l'écran.
    // Positionierung des Cursors zum Schreiben auf dem Bildschirm.
    // Positioning the cursor to write on the screen.
    Display.setCursor(0,12);

    // Affichage du mot "Hello !" à l'écran.
    // Anzeige des Wortes „Hallo!“ auf dem Bildschirm.
    // Display of the word “Hello!” on the screen.
    Display.println(F("Hello !"));


    // Initialisation de la radio LoRa.
    // Initialisieren des LoRa-Radios.
    // Initializing the LoRa radio.
    RfPort.begin();
    Radio = new Module(LORA_CS, LORA_DIO1, LORA_RST, LORA_BUSY, RfPort, SpiSettings);
 
    if(!Radio.begin(LORA_FREQUENCY, LORA_SIGNAL_BANDWIDTH, LORA_SPREADING_FACTOR, LORA_CODING_RATE_4, LORA_SYNC_WORD, LORA_TX_POWER_DBM))
    {
      // Attachement de la fonction d'interruption pour la réception radio LoRa.
      // Unterbrechungsfunktion für LoRa-Funkempfang anhängen.
      // Attach interrupt function for LoRa radio reception.
      Radio.setDio1Action(SetRxFlag);

      // Lancement de la réception radio LoRa.
      // Start des LoRa-Funkempfangs.
      // Launch of LoRa radio reception.
      if(!Radio.startReceive()) Display.println(F("LoRa OK"));
      else Display.println(F("LoRa KO"));
    }
    else Display.println(F("LoRa KO"));


    // Mise à jour de l'écran pour prendre en compte les actions ci-dessus.
    // Der Bildschirm wurde aktualisiert, um die oben genannten Aktionen zu berücksichtigen.
    // Updated the screen to take into account the actions above.
    Display.update();
}


/**** DEFINITION DE LA FONCTION D'EXECUTION DU MODULE LILYGO T-ECHO ****/
/**** DEFINITION DER AUSFÜHRUNGSFUNKTION DES LILYGO T-ECHO-MODULS ****/
/**** DEFINITION OF THE EXECUTION FUNCTION OF THE LILYGO T-ECHO MODULE ****/
void loop()
{
  // Réception d'une trame radio LoRa.
  // Empfang eines LoRa-Funkrahmens.
  // Reception of a LoRa radio frame.
  if(Receive)
  {
    // Lecture de la trame.
    // Den Frame lesen.
    // Read the frame.
    State = Radio.readData(TNC2Frame);

    // Effacement de l'écran et positionnement du curseur.
    // Bildschirm löschen und Cursor positionieren. 
    // Clearing the screen and positioning the cursor.
    Display.fillScreen(GxEPD_WHITE);
    Display.setCursor(0,12);

    // Affichage de la trame, du RSSI et du SNR à l'écran.
    // Anzeige von Frame, RSSI und SNR auf dem Bildschirm.
    // Displaying frame, RSSI and SNR on screen. 
    Display.println(F("**** RX ****"));
 
    if(State == ERR_NONE && TNC2Frame.charAt(0) == LORA_START_BYTE_1 && TNC2Frame.charAt(1) == LORA_START_BYTE_2 && TNC2Frame.charAt(2) == LORA_START_BYTE_3)
    {
      Display.println(TNC2Frame.substring(3));
  
      Display.print(F("RSSI:\t"));
      Display.print(Radio.getRSSI());
      Display.println(F(" dBm"));
  
      Display.print(F("SNR:\t"));
      Display.print(Radio.getSNR());
      Display.println(F(" dB"));
    }

    // Gestion des erreurs de réception.
    // Management von Empfangsfehlern.
    // Management of reception errors.
    else if(State == ERR_RX_TIMEOUT) Display.println(F("TimeOut !"));

    else if(State == ERR_CRC_MISMATCH) Display.println(F("CRC Error !"));

    else
    {
      Display.print(F("Failed : "));
      Display.println(State);
    }

    // Initialisation de l'indicateur de réception.
    // Initialisierung der Empfangsanzeige.
    // Initialization of the reception indicator.
    Receive = false;
  
    // Lancement de la réception radio LoRa.
    // Start des LoRa-Funkempfangs.
    // Launch of LoRa radio reception.
    State = Radio.startReceive();

    if(State != ERR_NONE)
    {
      Display.fillScreen(GxEPD_WHITE);
      Display.setCursor(0,12);

      Display.println(F("- Start Receive -"));
      Display.print(F("Failed : "));
      Display.println(State);
    }
    
    // Mise à jour de l'écran pour prendre en compte les actions ci-dessus.
    // Der Bildschirm wurde aktualisiert, um die oben genannten Aktionen zu berücksichtigen.
    // Updated the screen to take into account the actions above.
    Display.update();
  }
 
  
  // Rétroéclairage de l'écran.
  // Bildschirmhintergrundbeleuchtung.
  // Screen backlight.
  if(digitalRead(USER_TOUCH) == LOW) digitalWrite(E_PAPER_BACKLIGHT, HIGH);
  else digitalWrite(E_PAPER_BACKLIGHT, LOW);

  // Arrêt.
  // Stoppen.
  // Power off.
  if((digitalRead(USER_BUTTON) == LOW) && PowerEnable)
  {
      Display.fillScreen(GxEPD_WHITE);
      Display.update();

      PowerEnable = false;
      digitalWrite(POWER_ENABLE, LOW);

      while(digitalRead(USER_BUTTON) == LOW);
  }
  
  // Mise en marche.
  // Ein.
  // Power on.
  else if((digitalRead(USER_BUTTON) == LOW) && !PowerEnable)
  {
      digitalWrite(POWER_ENABLE, HIGH);
      PowerEnable = true;

      Display.fillScreen(GxEPD_WHITE);
      Display.setCursor(0,12);
      Display.println(F("POWER ON"));
      Display.update();
      
      while(digitalRead(USER_BUTTON) == LOW);
  }
}
